/*
Copyright (C) 2009 Frédéric Zubler, Rodney J. Douglas,
Dennis Göhlsdorf, Toby Weston, Andreas Hauri, Roman Bauer,
Sabina Pfister & Adrian M. Whatley.

This file is part of CX3D.

CX3D is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CX3D is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CX3D.  If not, see <http://www.gnu.org/licenses/>.
*/


package ini.cx3d.simulations.frontiers;


import static ini.cx3d.utilities.Matrix.randomNoise;
import static ini.cx3d.utilities.Matrix.subtract;
import ini.cx3d.Param;
import ini.cx3d.physics.ECMChemicalReaction;
import ini.cx3d.physics.PhysicalNode;
import ini.cx3d.physics.Substance;
import ini.cx3d.simulations.ECM;
import ini.cx3d.simulations.Scheduler;
import ini.cx3d.spatialOrganization.PositionNotAllowedException;

/**
 * This class was used to produce Figure 3 A-F of the paper
 * "A framework for modeling the growth and development of neurons and networks", Zubler & Douglas 2009.
 * 
 * It outputs a MATLAB script for generating the concentration plots.
 * 
 * @author fredericzubler
 *
 */
public class Figure_3_AF {


	// ECM
	public static ECM ecm = ECM.getInstance();

	// three template Substances
	static Substance templateA = new Substance("A", 1000,0.00);
	static Substance templateB = new Substance("B", 1000,0.00);
	static Substance templateC = new Substance("C", 1000,0.00);


	// ****************************************************************************************
	// ** MAIN
	// ****************************************************************************************

	public static void main(String[] args) {

		ecm.addNewSubstanceTemplate(templateA);
		ecm.addNewSubstanceTemplate(templateB);
		ecm.addNewSubstanceTemplate(templateC);
		
		//uncomment for figure ABC:
		//	testAddingRemovingNodes();
		//----------------------------------------
		// uncomment for figure D-E:
		testChemicalReaction();
		//-----------------------------------------
		// uncomment for figure F:
		// testMovingNodes();

	}


	// ****************************************************************************************
	// ** BENCH TESTS
	// ****************************************************************************************

	

	/*
	 * Figure 3 ABC:
	 * Pure diffusion over 500 nodes
	 */
	private void testDiffusion(){
		// 1) add  Physical Nodes
		addNodesAndGradient(100);

		// 2) Compute total quantity
		double[] quant = getTotalAmountABC();
				System.out.println("%Total quantity A before move = "+quant[0]);
				System.out.println("%Total quantity B before move = "+quant[1]);
				System.out.println("%Total quantity C before move = "+quant[2]);

		// 3) Diffuse a bit
		
		Scheduler.setPrintCurrentECMTime(false);
		int soManyTimeSteps = 50;
		for (int i = 0; i < soManyTimeSteps; i++) {
			Scheduler.simulateOneStep();
		}
		// Print total quantities after add/remove
		double[] quantafter = getTotalAmountABC();
		System.out.println("%Total quantity A after move = "+quantafter[0]);
		System.out.println("%Total quantity B after move = "+quantafter[1]);
		System.out.println("%Total quantity C after move = "+quantafter[2]);

		// 4) print concentration after
		printConcentrations();
	}


	/*
	 * Figure 3 DE
	 * Chemical reaction of 500 nodes
	 */
	private static void testChemicalReaction(){
		// add nodes
		addNodesAndGradient(100);
		// instantiate a chemical reaction
		Reac reac = new Reac();
		// run the thing
		Scheduler.setPrintCurrentECMTime(false);
		for (int i = 0; i < 1000; i++) {
			for (PhysicalNode pn: ECM.getInstance().getPhysicalNodeList()) {
				reac.run(pn);

			}
			Scheduler.simulateOneStep(); // comment for suppressing diffusion
		}
		// print results
		printConcentrations();
	}



	/* 
	 * Figure 3F:
	 * Add a bunch of points, move them and looking at the concentration before and after.
	 */
	private void testMovingNodes(){

		
		// define 3 special nodes that will be moved 
		double[] intialPn1Position = {0,0,0.7};
		double[] intialPn2Position = {-1,-3,4};
		double[] intialPn3Position = {6,2,-10};

		PhysicalNode pn1 = ecm.getPhysicalNodeInstance(intialPn1Position.clone());
		PhysicalNode pn2 = ecm.getPhysicalNodeInstance(intialPn2Position.clone());
		PhysicalNode pn3 = ecm.getPhysicalNodeInstance(intialPn3Position.clone());

		// add 500 others, and fill them all with substances
		addNodesAndGradient(100);

		// ask for total quantity before movements
		double[] quant = getTotalAmountABC();
		System.out.println("%Total quantity A before move = "+quant[0]);
		System.out.println("%Total quantity B before move = "+quant[1]);
		System.out.println("%Total quantity C before move = "+quant[2]);


		// move a few times the 3 special nodes pn1, pn2, pn3
		int numberOfMoves = 100;
		for (int i = 0; i < numberOfMoves; i++) {

			double[] dX1 = randomNoise(5.0,3);
			double[] dX2 = randomNoise(5.0,3);
			double[] dX3 = randomNoise(5.0,3);

			if(i%15==1){
				dX1 = subtract(intialPn1Position, pn1.getSoNode().getPosition());
			}
			if(i%15==2){
				dX2 = subtract(intialPn2Position, pn2.getSoNode().getPosition());
			}
			if(i%15==3){
				dX3 = subtract(intialPn3Position, pn3.getSoNode().getPosition());
			}


			try{
				pn1.getSoNode().moveFrom(dX1);
				pn2.getSoNode().moveFrom(dX2);
				pn3.getSoNode().moveFrom(dX3);
			} catch (PositionNotAllowedException e) {
				e.printStackTrace();
			}		
		}


		//		ask for total quantity after movement
		double[] quantafter = getTotalAmountABC();
		System.out.println("%Total quantity A after move = "+quantafter[0]);
		System.out.println("%Total quantity B after move = "+quantafter[1]);
		System.out.println("%Total quantity C after move = "+quantafter[2]);

		printConcentrations();
	}

	


	/*
	 * This inner class represents a chemical reaction, and therefore implements the
	 * ECMChemicalReaction interface. It is used for Figure D and E
	 */
	private static class Reac implements ECMChemicalReaction{

		static double k = 10*3;
		static double k_1 = 0.5*3;

		public void run(PhysicalNode pn) {

			double volume = pn.getSoNode().getVolume();
			double R = pn.getExtracellularConcentration("A");
			double G = pn.getExtracellularConcentration("B");
			double B = pn.getExtracellularConcentration("C");


			// R+G <==> B
			double dR_dt = -k*R*G + k_1*B;

			pn.getSubstanceInstance(templateA).setConcentration(R+dR_dt*Param.SIMULATION_TIME_STEP);
			pn.getSubstanceInstance(templateA).updateQuantityBasedOnConcentration(volume);
			pn.getSubstanceInstance(templateB).setConcentration(G+dR_dt*Param.SIMULATION_TIME_STEP);
			pn.getSubstanceInstance(templateB).updateQuantityBasedOnConcentration(volume);
			pn.getSubstanceInstance(templateC).setConcentration(B-dR_dt*Param.SIMULATION_TIME_STEP);
			pn.getSubstanceInstance(templateC).updateQuantityBasedOnConcentration(volume);

		}

	}
	
	

	// ****************************************************************************************
	// ** UTIL
	// ****************************************************************************************

	private static void printConcentrations(){
		System.out.println("b = [");
		for (int i = 0; i < ECM.getInstance().physicalNodeList.size(); i++) {
			PhysicalNode tempN = ECM.getInstance().physicalNodeList.get(i);
			double position = tempN.getSoNode().getPosition()[2];	// z value
			double quantityA = tempN.getExtracellularConcentration("A");
			double quantityB = tempN.getExtracellularConcentration("B");
			double quantityC = tempN.getExtracellularConcentration("C");
			System.out.println(position+" , "+
					quantityA+" , "+quantityB+" , "+quantityC);
		}
		System.out.println("];");
		System.out.println("plot(b(:,1),b(:,2),'.r'), hold on; plot(b(:,1),b(:,3),'.g'), plot(b(:,1),b(:,4),'.b')");
	}

	
	/*
	 * 500 physical nodes, in volume [-size ; +size]^3 , with three chemicals
	 */
	private static void addNodesAndGradient(double size){

		// add empty PhysicalNodes
		for (int i = 0; i < randomCoord.length; i++) {
			ecm.getPhysicalNodeInstance(randomCoord[i]);
		}


		// add chemical inside the PhysicalNodes
		for (PhysicalNode n : ecm.getPhysicalNodeList()) {

			double volume = n.getSoNode().getVolume();
			double desiredConcntration; 

			// Substance A:  uniformly distributed
			if(n.getSoNode().getPosition()[2]>0){
				desiredConcntration = 0.2;				
			}else{
				desiredConcntration = 1.0;
			}
			n.modifyExtracellularQuantity("A", volume*desiredConcntration/Param.SIMULATION_TIME_STEP);

			// B is a vertical linear gradient
			desiredConcntration = 0.8 + n.getSoNode().getPosition()[2]/(2*size);
			n.modifyExtracellularQuantity("B", volume*desiredConcntration/Param.SIMULATION_TIME_STEP);

			// C is a cosine
			desiredConcntration = 1.0 + 0.8*Math.cos(n.getSoNode().getPosition()[2]*6.28/(2*size));
			n.modifyExtracellularQuantity("C", volume*desiredConcntration/Param.SIMULATION_TIME_STEP);
		}
	}




	/* Summation of the total amount of the three chemicals in the whole space.*/
	private double[] getTotalAmountABC(){
		double totalQuantityA = 0;
		double totalQuantityB = 0;
		double totalQuantityC = 0;

		for (PhysicalNode tempN: ECM.getInstance().getPhysicalNodeList()) {
			double volume = tempN.getSoNode().getVolume();

			totalQuantityA += tempN.getExtracellularConcentration("A")*volume;
			totalQuantityB += tempN.getExtracellularConcentration("B")*volume;
			totalQuantityC += tempN.getExtracellularConcentration("C")*volume;
		}
		return new double[] {totalQuantityA, totalQuantityB, totalQuantityC};
	}


	// The nodes position (so that we use the same positions for each test)
	public static double[][] randomCoord = 
	{
		{31.70302, -72.26469, -70.3131}, 
		{51.922436, -63.867596, 61.573715}, 
		{-91.09586, 67.12055, -99.24571}, 
		{52.4107, 68.65413, -58.918533}, 
		{53.236706, -21.825605, -4.3315763}, 
		{-71.52755, -41.416874, -29.95678}, 
		{99.95046, 76.57273, 80.544029}, 
		{91.71856, -36.21893, 12.469306}, 
		{56.007908, -70.935684, -11.188955}, 
		{82.224396, -6.223608, 63.2154}, 
		{67.55541, 20.682777, -31.489288}, 
		{75.452095, 61.112026, -41.09262}, 
		{-85.730255, -36.417213, -33.315346}, 
		{60.80221, 93.942604, 97.27286}, 
		{-93.50118, 21.301538, -88.98847}, 
		{-96.2973, 43.956295, 95.574936}, 
		{78.46376, 55.133263, -86.618}, 
		{62.01791, -73.6963, 26.402905}, 
		{36.474876, -48.631813, 33.999718}, 
		{-96.92566, -28.417833, -99.89189}, 
		{85.33251, -70.10379, -98.825775}, 
		{-91.61619, 60.49499, 94.43872}, 
		{8.404943, -79.00781, 73.926266}, 
		{42.208416, 65.40054, -95.44291}, 
		{36.245274, -82.75778, -46.07256}, 
		{64.83622, 14.782306, -83.87365}, 
		{63.34755, 35.17089, 89.58954}, 
		{93.58104, -88.722626, -29.589207}, 
		{-49.55549, 64.197105, 9.124884}, 
		{40.216377, 15.243236, -92.394035}, 
		{77.686104, 31.461792, -28.687475}, 
		{3.3098783, 7.241096, 41.53157}, 
		{-51.446148, 96.59455, 52.096863}, 
		{-27.375437, -23.036049, 28.3915}, 
		{51.282433, -98.25101, 60.960083}, 
		{49.652885, -24.8906, 1.9128088}, 
		{-10.929186, 84.74703, 11.568664}, 
		{27.083654, 89.93587, -91.4222}, 
		{-72.95768, 46.81347, -55.808502}, 
		{-41.423374, -37.432804, -74.00744}, 
		{-33.299435, 15.411162, 59.89433}, 
		{31.01875, -15.280575, 55.19988}, 
		{-49.153217, 5.7371426, 54.976433}, 
		{1.2171762, -11.01437, -1.1733203}, 
		{-13.730284, 84.519, -45.648003}, 
		{-61.120285, -35.92522, 60.61416}, 
		{20.060818, -55.758686, -90.214905}, 
		{23.251951, 33.779743, -98.254745}, 
		{-50.40805, 6.59191, 37.634388}, 
		{51.803925, 16.642284, -26.53311}, 
		{-73.160866, 76.87981, -35.00357}, 
		{-49.854004, -69.63982, -93.813614}, 
		{35.040134, -13.050888, -13.824175}, 
		{21.536695, 77.38832, 99.26019}, 
		{79.20605, 77.57805, 15.950731}, 
		{-25.068222, 11.742252, -65.778206}, 
		{10.712402, 21.768328, 84.21484}, 
		{-60.102753, 66.91477, 72.72166}, 
		{-21.605167, -98.67081, 29.706093}, 
		{-33.37332, -44.3668, 28.858866}, 
		{1.962406, 78.75114, -90.500496}, 
		{55.898376, 27.86351, 88.12842}, 
		{45.1892, 96.92903, -4.0592055}, 
		{37.587727, -2.6731555, 30.524273}, 
		{-67.30312, 67.59109, 43.300167}, 
		{72.19827, -18.470285, 86.43239}, 
		{-78.982086, -63.548603, -8.423416}, 
		{77.04847, 17.991941, -21.79556}, 
		{-22.77365, -65.362144, -37.09778}, 
		{-49.11672, -48.568966, 79.47818}, 
		{-78.689186, -78.26347, -15.4194}, 
		{-4.2218056, -26.051018, 44.53852}, 
		{-4.4122777, -89.90389, 39.009384}, 
		{-94.34928, -57.037903, 41.800926}, 
		{89.13288, 31.484646, -48.963444}, 
		{33.24896, -90.71239, 73.7299}, 
		{80.14026, 98.79291, -56.50435}, 
		{-26.718912, -70.603134, -80.56372}, 
		{86.37025, 44.617855, 46.958527}, 
		{-46.108784, -10.315231, 19.462431}, 
		{14.869253, 36.993053, -25.298008}, 
		{41.85947, 74.38413, 92.87033}, 
		{-32.24965, 16.39331, 62.453564}, 
		{-97.03576, 9.541551, -22.921947}, 
		{41.214584, -62.063194, 27.981672}, 
		{66.52609, -80.28582, 87.70243}, 
		{-90.26412, 93.32122, -9.277868}, 
		{-94.737274, 57.564453, 70.09264}, 
		{35.374012, -82.280106, -21.544155}, 
		{25.606424, 84.51751, 51.78918}, 
		{11.840196, -68.73851, 68.01827}, 
		{71.32631, 77.67455, -14.302594}, 
		{31.914116, -69.16897, -99.05114}, 
		{-39.20855, -41.810616, -88.416664}, 
		{-47.830364, 28.142967, 83.07396}, 
		{-73.98588, -67.36398, 57.213203}, 
		{-22.845703, 51.107372, 28.830675}, 
		{4.609088, -75.78339, -16.923628}, 
		{-37.994095, 98.569885, -84.37467}, 
		{93.25167, 67.05309, 63.535797}, 
		{63.172348, -90.9362, 49.001423}, 
		{26.748022, 25.715982, -55.800842}, 
		{85.99966, 2.1500177, -56.439186}, 
		{58.452816, -25.607492, -43.56993}, 
		{-20.587233, -59.502937, 44.510994}, 
		{27.865576, 75.87643, 75.73014}, 
		{15.054036, 90.69974, -43.155533}, 
		{86.56477, -0.5537984, -63.898296}, 
		{-76.23822, 13.961338, -70.68548}, 
		{75.33147, 27.629955, -53.29397}, 
		{-81.76507, -38.448765, -66.249794}, 
		{-1.1205095, -46.843143, 73.795166}, 
		{38.58594, -13.843192, -28.577862}, 
		{9.929753, -53.24106, 58.02509}, 
		{58.05239, 95.00755, 71.07576}, 
		{5.7000375, -19.013262, 27.236958}, 
		{79.61397, -15.531565, -28.790493}, 
		{-44.138603, -18.71942, -35.621468}, 
		{-49.784515, -86.43854, 14.887824}, 
		{-45.35413, -90.981964, -89.43629}, 
		{-13.325799, -65.16274, 50.913208}, 
		{38.46036, 87.3317, 69.840935}, 
		{24.470732, 28.814384, -88.017296}, 
		{-14.1737, 93.73079, 64.94466}, 
		{5.183007, 40.841896, 19.342985}, 
		{54.856365, -42.089756, 36.395664}, 
		{-42.607975, 39.10199, 94.629654}, 
		{85.30671, 21.91218, -70.57135}, 
		{-94.08782, -23.686735, -97.39932}, 
		{-47.612236, 5.7123804, -42.633217}, 
		{80.99776, 6.6534023, -77.70316}, 
		{-2.7626905, -53.59103, 33.616}, 
		{67.94622, 79.00082, 77.10339}, 
		{47.88835, -25.062881, -70.82571}, 
		{-77.171295, 77.68117, -66.0574}, 
		{-7.290212, -98.557526, -75.42418}, 
		{-74.12678, 26.75391, -94.53363}, 
		{-31.95785, -0.6869906, 5.0245767}, 
		{3.6280823, 12.119543, 13.087758}, 
		{99.852135, -82.036606, -2.1178543E-4}, 
		{-62.864178, -59.622704, 58.80881}, 
		{-93.17939, 84.734436, -66.906364}, 
		{-91.9772, 61.798992, -64.73278}, 
		{-25.132347, -44.968372, -52.0963}, 
		{-0.89835346, -11.235117, 10.803578}, 
		{-19.413073, 74.7777, -30.544409}, 
		{-48.386948, 76.66494, 53.59214}, 
		{37.669155, -22.536009, -9.741663}, 
		{-62.204746, -27.888573, -90.70016}, 
		{14.772516, -43.99993, 12.45464}, 
		{94.63498, -71.78065, -16.963411}, 
		{8.454815, -56.999474, -6.248496}, 
		{-98.633354, 27.946978, -29.394907}, 
		{18.100983, 77.30754, -69.586365}, 
		{97.86897, -82.31612, -32.304317}, 
		{-36.98256, 13.998016, -54.54638}, 
		{21.25403, -57.565754, 99.52123}, 
		{91.155235, 93.8102, 48.51798}, 
		{80.72407, -16.546335, -42.355328}, 
		{74.77421, 79.16698, 94.47922}, 
		{-47.368538, -44.18555, 14.986872}, 
		{-24.346308, 0.4664834, -76.59124}, 
		{24.436712, 38.23115, 83.16154}, 
		{-69.55079, 60.126415, 32.80727}, 
		{73.11472, 26.082403, -43.313976}, 
		{79.12194, 94.45084, -34.535614}, 
		{-34.920162, 50.799755, -6.0416923}, 
		{94.03661, 85.55819, -70.75646}, 
		{4.8316717, 58.91698, 62.241108}, 
		{6.7163, 97.56189, -31.054792}, 
		{41.586246, -59.282703, 29.69917}, 
		{50.117207, 42.500877, 40.078304}, 
		{-70.15401, 24.588392, -62.384045}, 
		{-2.2236547, -18.87909, -20.878872}, 
		{3.247787, -23.720411, 5.4663177}, 
		{-17.20311, 59.830753, 98.31335}, 
		{92.55631, -35.990974, -4.8875737}, 
		{-0.74650097, 65.69361, -36.80271}, 
		{39.160023, -43.075184, -68.032814}, 
		{40.4806, -56.76478, 30.402254}, 
		{-74.92479, 36.289932, 3.851917}, 
		{11.48326, -83.39936, 58.29833}, 
		{68.8395, 20.819729, 45.159595}, 
		{-0.6094132, 63.991978, 65.26841}, 
		{-7.274422, -20.889744, -13.109007}, 
		{13.838931, 23.565197, -55.03289}, 
		{-71.251175, 46.68882, 93.604126}, 
		{-29.816631, -94.37501, 84.93547}, 
		{36.674667, 40.902405, -92.87187}, 
		{18.624956, 38.97909, -61.421085}, 
		{3.722456, -83.9818, -75.76409}, 
		{-7.7851233, -6.9838934, 19.42523}, 
		{-10.641175, 0.3008545, -89.08374}, 
		{11.138368, 54.15109, -96.232315}, 
		{-14.706103, -32.413975, 18.380714}, 
		{28.948677, 67.808914, 10.648365}, 
		{-44.265133, -86.50202, 43.803894}, 
		{-87.85289, 16.534494, 78.40591}, 
		{-9.317074, 10.841574, -40.27685}, 
		{15.69745, -68.05055, -39.166393}, 
		{-10.4732895, -75.90453, -82.91491}, 
		{46.10432, -66.68645, -56.124657}, 
		{46.709095, 67.97579, 64.56753}, 
		{17.333216, -42.034023, -59.38868}, 
		{48.51888, 32.735287, -13.15342}, 
		{-85.76199, -32.138847, -2.689155}, 
		{-67.844185, -13.404154, -27.885206}, 
		{28.797146, 60.097927, -39.81216}, 
		{-13.4188795, -44.76053, -91.04192}, 
		{-43.774208, -84.536964, 69.19509}, 
		{-19.88316, 5.7369795, 78.61171}, 
		{95.18507, -65.35803, 4.117037}, 
		{32.491325, 98.18325, 53.959633}, 
		{37.484814, 43.336685, 86.12707}, 
		{68.61512, 67.515175, -95.88869}, 
		{40.88007, 10.890935, -4.003847}, 
		{-83.24331, 49.77138, -94.942314}, 
		{21.177015, 97.3557, -96.21723}, 
		{73.29152, -54.9557, 4.21651}, 
		{-21.350988, 5.7990975, 93.98302}, 
		{37.51587, 21.582972, -79.55065}, 
		{-25.14934, 14.023358, 29.031603}, 
		{-79.781204, -95.68809, 81.56132}, 
		{-80.18504, 46.015224, 21.35221}, 
		{5.2376833, -16.804886, -82.04789}, 
		{-78.79118, -68.87801, -87.13241}, 
		{59.904232, 33.528046, -87.08257}, 
		{25.562902, 10.182532, 8.3780365}, 
		{-26.685137, 6.891345, -64.5758}, 
		{-29.183731, 64.09727, 27.259424}, 
		{85.30174, -41.968163, -12.303893}, 
		{-72.86847, 52.412537, -37.347572}, 
		{-52.21611, -16.39413, 34.71846}, 
		{-28.3736, -92.38015, -60.780922}, 
		{-62.837498, 12.160817, -47.751312}, 
		{99.088905, -5.402405, 94.51625}, 
		{-25.714968, -16.049871, 49.920403}, 
		{76.02986, 32.899036, 64.333824}, 
		{79.31069, 61.55838, 71.75148}, 
		{95.08555, 20.73476, 2.60282}, 
		{-75.99057, -80.38664, 32.379154}, 
		{60.697388, 81.20641, -80.2982}, 
		{-15.561205, 60.21505, -42.40623}, 
		{-50.576637, 70.69133, -28.437292}, 
		{-57.23722, 10.624272, -84.45252}, 
		{37.80916, 2.699942, -97.34336}, 
		{47.77482, 86.10941, -36.929523}, 
		{10.474446, -86.14962, -50.140816}, 
		{88.40811, 82.90617, -88.09484}, 
		{-31.666334, -42.978077, 96.862}, 
		{41.48153, 27.63514, -76.18621}, 
		{-96.57097, 84.83224, -39.905884}, 
		{61.832996, -91.74675, -37.752342}, 
		{-71.16087, -21.722963, -48.697903}, 
		{36.905594, 90.070564, -88.21917}, 
		{20.047083, 83.9959, -52.355717}, 
		{35.21908, 55.300995, -15.620105}, 
		{-7.1089973, -88.120476, 90.38965}, 
		{-45.19763, 57.167812, -17.252525}, 
		{-97.96658, 84.68188, 95.45845}, 
		{-1.224929, 45.259106, -86.237274}, 
		{66.40934, 54.39116, -5.141953}, 
		{-44.580086, -22.491358, -72.85912}, 
		{-77.648415, 47.166103, 23.685873}, 
		{-16.773376, 0.12043002, -7.082462}, 
		{63.430058, -30.1451, 43.299736}, 
		{-49.166462, 48.594566, -17.323545}, 
		{-24.763046, 16.05214, 52.37725}, 
		{-29.760141, 33.754982, 49.402203}, 
		{24.534578, -82.738525, 23.158245}, 
		{11.180071, -56.29096, -65.35235}, 
		{60.864857, 73.336754, 33.819263}, 
		{76.58772, -32.256813, 3.9294596}, 
		{-35.217464, 3.669243, -92.32347}, 
		{-55.55111, 67.40995, 15.849626}, 
		{94.57485, -66.3581, -41.865566}, 
		{90.251335, -46.002056, -29.293505}, 
		{-8.697474, -23.869581, 51.749573}, 
		{56.24437, 35.080765, -76.99901}, 
		{-16.752365, -98.88636, 1.5981652}, 
		{47.781303, 55.17041, 51.33255}, 
		{-47.23366, 82.9595, -15.175344}, 
		{41.442524, 50.2893, 90.16036}, 
		{-7.1237493, 59.24681, -78.30486}, 
		{72.12675, -28.470312, 81.59883}, 
		{-26.20775, 71.01266, 72.93782}, 
		{-76.077866, 0.7729263, -97.98269}, 
		{81.56874, -64.59845, -93.74237}, 
		{-61.227333, -90.548164, -32.70678}, 
		{86.89334, -59.47045, 84.670555}, 
		{-56.441357, 9.689911, -8.929684}, 
		{-8.835474, 71.03601, -43.901382}, 
		{5.4089246, 61.849117, 40.619736}, 
		{2.9072053, -47.27809, -72.74415}, 
		{91.42688, 63.346386, -57.413185}, 
		{78.34232, -63.31285, 21.372747}, 
		{26.487152, -85.18443, 83.10077}, 
		{57.91356, -22.32392, -80.74172}, 
		{14.666441, -10.070877, 29.873537}, 
		{-77.45756, 57.48947, 21.46874}, 
		{-21.09542, 7.4602838, 93.83775}, 
		{-59.53338, -18.730652, -64.80212}, 
		{51.721382, 85.844345, 19.767971}, 
		{73.221886, 60.52924, -43.44309}, 
		{-51.31347, 53.34524, -66.236115}, 
		{-62.460384, 79.68296, -4.816818}, 
		{16.673769, 77.0833, 24.394665}, 
		{92.9059, 19.589691, -34.01405}, 
		{-29.077673, 82.13965, 47.03018}, 
		{-59.10118, -61.272675, -20.936377}, 
		{28.148327, 62.60557, 30.112896}, 
		{-84.805275, -44.14965, 4.5433846}, 
		{76.34201, 79.04827, -50.07591}, 
		{-72.30501, -37.422222, -80.19255}, 
		{-26.2267, 15.051322, -91.78502}, 
		{68.33009, 54.861618, -7.835631}, 
		{77.80411, 87.26631, 85.72761}, 
		{69.88043, -97.34781, -25.24399}, 
		{-84.55953, -88.83586, 4.31746}, 
		{-84.87932, 80.62338, 45.993923}, 
		{28.597588, 95.15897, -29.320232}, 
		{36.630035, 22.244349, 86.101006}, 
		{73.615, 30.126387, 14.520007}, 
		{-71.028244, 6.8544083, -74.22214}, 
		{1.8438632, 55.058937, -98.28405}, 
		{32.265152, -23.84005, -76.62364}, 
		{-19.296198, -7.4207788, -13.710125}, 
		{-36.541157, 43.546944, -85.62181}, 
		{76.41749, 74.520325, 60.91465}, 
		{59.24939, -93.01506, -49.048214}, 
		{-36.001442, -83.58015, -69.337524}, 
		{79.331245, 13.658833, 14.778769}, 
		{-30.312828, -16.43674, 34.589733}, 
		{11.645719, -79.25154, -17.539455}, 
		{-92.03224, 37.538902, 99.73144}, 
		{-31.974035, 30.303274, -26.732279}, 
		{31.01137, -87.11197, -33.275623}, 
		{88.99457, 84.62593, -28.211143}, 
		{53.67966, 84.48092, 69.534355}, 
		{-45.28172, 39.054108, 35.553013}, 
		{-97.69444, 18.233713, -13.049552}, 
		{73.89304, -12.039976, 46.700153}, 
		{19.895945, 5.5816474, -26.676718}, 
		{-23.369574, 22.992397, -91.6118}, 
		{63.131218, 55.000763, -7.464972}, 
		{38.62548, -74.739265, -83.63759}, 
		{75.88373, -74.54423, -46.15829}, 
		{13.765024, -80.12216, -66.144516}, 
		{-92.36073, -55.04685, -78.15638}, 
		{37.016087, 95.978004, -22.007862}, 
		{-39.748688, 54.779163, -54.387203}, 
		{-0.5745829, -28.663689, -13.226246}, 
		{-50.666527, -16.739786, -66.492775}, 
		{97.24461, 70.72358, -89.47588}, 
		{23.120932, 93.13016, 59.376144}, 
		{56.830273, -86.78713, 77.76924}, 
		{17.002295, -90.66576, 1.0284611}, 
		{-40.313713, -52.748688, 74.6609}, 
		{-29.290504, 49.224472, -79.306465}, 
		{-4.1749372, -75.75306, 85.36713}, 
		{91.30031, -91.13519, -53.180325}, 
		{-99.09233, -88.11189, 38.851852}, 
		{11.89626, -43.80112, -86.62341}, 
		{-39.86147, -23.015205, 50.03247}, 
		{-63.656708, -72.85364, 72.673706}, 
		{50.770298, -66.75954, 19.696484}, 
		{-54.38599, 81.66889, -4.227691}, 
		{-59.172657, -19.648378, 41.40929}, 
		{-95.6863, 55.240753, 95.425865}, 
		{44.164337, 44.884346, -37.03771}, 
		{85.90832, -76.89381, -5.924836}, 
		{-34.50233, 67.56384, -42.923176}, 
		{33.456245, 0.7117756, 67.073555}, 
		{52.214787, -77.62103, -2.1473846}, 
		{98.267105, -61.348297, 31.717247}, 
		{-64.41251, -33.93018, -73.037544}, 
		{20.976068, -26.385815, -52.122406}, 
		{9.873038, 27.327913, 72.35531}, 
		{16.330664, -8.1993265, 27.520855}, 
		{-84.56102, -12.035091, 82.12096}, 
		{32.23459, -93.88529, -91.022675}, 
		{-2.3729837, 77.343765, -78.55164}, 
		{-77.64998, 85.46623, -12.319443}, 
		{-22.914429, -0.1650764, 99.91009}, 
		{6.2713366, -19.431778, 67.80923}, 
		{44.633293, 5.655223, -67.36615}, 
		{7.76942, -73.25154, -72.2252}, 
		{73.62333, -29.760529, -40.337044}, 
		{-64.93749, -33.932735, 5.126673}, 
		{98.895, 97.72859, -45.282867}, 
		{-17.933609, -1.6938933, 90.88974}, 
		{51.732166, -48.945232, -66.2223}, 
		{-69.94005, 57.428604, 76.67628}, 
		{-42.820423, 34.658306, 45.94305}, 
		{68.49088, -84.451515, 76.5709}, 
		{20.357128, 25.653187, -1.2844757}, 
		{-55.56278, -26.72425, 15.594017}, 
		{-36.8091, -65.7862, -23.453138}, 
		{-85.3975, -79.06454, -37.29895}, 
		{67.75566, 52.40467, -21.281696}, 
		{-6.8981757, 1.3216722, 23.342175}, 
		{67.080284, -39.686813, -11.527696}, 
		{53.51293, 7.4935613, 88.53549}, 
		{-36.937977, -53.75714, 32.127476}, 
		{4.9159884, 79.15128, 67.51604}, 
		{-7.2984123, 70.326614, 71.454956}, 
		{78.49826, -26.773203, 72.008835}, 
		{-8.784865, -25.95322, -43.994026}, 
		{52.700806, 67.748535, 44.106903}, 
		{-18.914328, -13.004599, 39.864132}, 
		{39.153923, 86.42094, 58.059895}, 
		{51.541836, 91.346535, 19.335506}, 
		{36.481518, -87.291626, 19.815956}, 
		{69.42987, -82.114586, -11.358055}, 
		{-68.47749, -11.089735, -62.895344}, 
		{0.48238885, -81.6306, 38.38168}, 
		{-55.93547, -5.4809933, -53.228695}, 
		{-75.53155, -99.82352, 69.3735}, 
		{-62.367126, -19.803713, -66.080536}, 
		{-62.10929, -57.29948, -20.110159}, 
		{-19.580385, 44.49525, 37.508255}, 
		{77.41036, -77.77813, -41.032574}, 
		{71.64646, -90.12187, -52.92015}, 
		{-75.96295, -56.159897, -81.40525}, 
		{-56.526684, -53.110775, 2.3307545}, 
		{4.700822, -34.304806, 93.62136}, 
		{-89.68531, -40.825417, -82.266914}, 
		{70.88603, 61.697037, 85.097565}, 
		{28.802813, -39.60035, -15.820243}, 
		{64.683655, 23.274632, -90.30782}, 
		{-75.29084, 14.388177, -76.13475}, 
		{76.41838, -23.433622, 18.483957}, 
		{-30.648117, -26.253376, 77.892006}, 
		{34.811657, -23.302849, -29.213766}, 
		{-11.805895, -74.018074, 32.35665}, 
		{64.710045, -67.60373, 85.12738}, 
		{58.673668, -39.82612, -8.759324}, 
		{97.372444, -31.351587, -33.169918}, 
		{-87.63161, -93.490234, -3.9980626}, 
		{57.177536, 18.15749, 40.941933}, 
		{-32.66228, 14.685962, 94.40691}, 
		{-6.232366, -30.663292, -53.92253}, 
		{-41.787823, 48.96183, 56.304424}, 
		{-68.156204, 23.886557, 98.91955}, 
		{-57.978474, 45.522865, 12.539442}, 
		{-91.79641, 24.732086, -15.046268}, 
		{-16.06885, 44.156334, 5.236621}, 
		{-72.32587, -91.51855, -82.893456}, 
		{-24.560314, -52.179688, 42.816795}, 
		{51.22206, -21.665218, 13.334056}, 
		{23.999063, 41.629818, 70.29197}, 
		{19.977156, 55.586456, 44.24476}, 
		{52.6465, -87.401115, 22.935053}, 
		{91.81964, 81.96966, 94.29732}, 
		{45.65716, 33.04806, -82.96445}, 
		{30.560751, 51.910675, 51.123158}, 
		{15.380794, 18.170574, 42.20121}, 
		{-66.02494, 53.486385, -92.407394}, 
		{-0.3632208, -59.519875, -88.0056}, 
		{-57.22122, 99.08251, 35.25563}, 
		{-58.995594, 13.746869, -30.579308}, 
		{49.223675, 0.37778744, -94.999306}, 
		{71.12935, -36.051544, -43.889397}, 
		{-33.200813, -34.349804, 14.82142}, 
		{6.218734, 36.4774, -26.475151}, 
		{61.489006, 8.211589, -6.5316367}, 
		{45.035767, 83.48812, 85.11374}, 
		{44.369633, 0.096341394, -75.12528}, 
		{-1.8736033, -96.656105, 33.392384}, 
		{-66.298805, 91.40368, 58.338276}, 
		{-22.114529, -23.844404, 11.415385}, 
		{28.515663, -12.377355, 28.395098}, 
		{-16.098583, -3.5162356, 21.772911}, 
		{-21.659145, 6.373126, -10.7421055}, 
		{0.45568255, 12.495708, -24.345268}, 
		{-3.7968438, 16.252483, 5.808786}, 
		{-6.0878797, -20.227003, 3.7969098}, 
		{-29.831146, 24.665112, 19.432964}, 
		{-2.1897619, 10.539925, -24.553045}, 
		{-9.511633, -29.89792, -17.035395}, 
		{17.97229, -17.069967, -25.292154}, 
		{17.156734, 18.764595, -25.154188}, 
		{10.862116, -28.224037, 7.1590314}, 
		{-28.745756, -24.090652, 12.723018}, 
		{-26.563574, 2.4002135, 24.530716}, 
		{3.915493, -12.44105, 5.611186}, 
		{-23.181793, -23.999897, -21.432049}, 
		{-8.253937, 25.754444, 11.328417}, 
		{16.16917, 1.0774037, 24.55905}, 
		{-26.730803, -10.629501, 4.131265}, 
		{-13.860291, -5.4484334, -17.747814}, 
		{0.5675854, 0.53845245, 22.829283}, 
		{-15.960665, -6.3824444, 11.790962}, 
		{2.027561, -20.38869, 8.903073}, 
		{-26.980846, -14.720091, 7.4225388}, 
		{24.646942, 17.85451, 29.383574}, 
		{24.524075, 25.341066, 7.442075}, 
		{-8.301709, 2.9650354, 2.8735151}, 
		{-4.697668, -1.7159088, -21.68201}, 
		{-12.076115, -2.7694888, 16.82006}, 
	};



}
